from langchain_openai import ChatOpenAI
from langchain.agents import AgentExecutor, create_openai_tools_agent
from langchain import hub
from langchain.tools import StructuredTool


class BlogAgent:
    def __init__(self, api_key=None) -> None:
        self.llm = ChatOpenAI(api_key=api_key)

        # blog_img_tool = generate_images_and_add_to_blog()
        blog_img_tool = StructuredTool.from_function(
            name="generate_images_and_add_to_blog", 
            func=generate_images_and_add_to_blog,
            description="This tool helps to generate images and add them to blog. Blog should contain image prompts in <image> prompt </image> tag."
        )
        self.tools = [blog_img_tool]

        self.llm = ChatOpenAI(api_key=api_key)#.bind_tools(self.tools)

    def write_blog_text(self, topic, context):
        prompt2 = (
            f"You are an effective Blog writer. Output only blog, without any additional text. "
            f"Write a comprehensive blog post based on the following details:\n\n"
            f"Topic: {topic}\n"
            f"Summarized context about the topic: {context}\n\n"
            f"The blog should include an engaging introduction to topic, then detailed stections about how the context addresses the topic, "
            f"and a conclusion summarizing the key points. Structure the blog with clear headings, and write it in a conversational style.  "
            f"Output the blog in markdown format, including a title, introduction, body sections, and conclusion. Write in a conversational style to engage readers. "
            f"Remember to add image description in '<image/>' tag where image is required(total 1 image only, <image>image description</image>). "
        )
        #f"Remember to use '<IMAGE/>' placeholder where image is required.(total 1 image only) "

        blog_text = self.llm.invoke(prompt2)
        return blog_text

    def generate_blog(self, topic, content):
        self.topic = topic
        blog_text = self.write_blog_text(topic, content)
        # blog_content = self.add_image_prompts(blog_text)
        # self.blog_content = blog_content
        # print(blog_content)
        # output = self.add_images(blog_content)
        # self.image_path = output[-1][-1][0]
        # doc_file = output[0][1]
        # return blog_text.replace("<-IMAGE->", "") , doc_file, self.image_path
        return blog_text
    
    # res = self.llm.invoke(prompt2)
    #     print(res.tool_calls[0])
    #     return generate_images_and_add_to_blog(**res.tool_calls[0]['args'])
    
def generate_images_and_add_to_blog(blog_content, save_temp=False):
    """This tool is used to generate images and add them to blog
    Args:
    blog_content: A complete blog with image prompts enclosed in <image> prompt </image> tag.
    Returns:
    A complete blog"""
    import tempfile
    import os
    import re
    import base64

    blog_content = str(blog_content)
    # print(f"****************\n{blog_content}\n**********")
    image_descriptions = re.findall(r'<image>(.*?)</image>', blog_content)
    
    md_file_path = 'blog_post.md'
    docx_file_path = 'blog_post.docx'
    if save_temp:
        temp_folder = tempfile.gettempdir()
        md_file_path = os.path.join(temp_folder, 'blog_post.md')
        docx_file_path = os.path.join(temp_folder, 'blog_post.docx')
    
    if os.path.exists(md_file_path):
        os.remove(md_file_path)
    if os.path.exists(docx_file_path):
        os.remove(docx_file_path)
    
    print(image_descriptions)
    images = []
    # base64_images = []
    for i, text in enumerate(image_descriptions):
        try:
            img_path = generate_image_openai(text,save_temp=save_temp, _i=i)
            print("image_generated")
            blog_content = blog_content.replace(f'<image>{text}</image>', f'![]({img_path})')
            images.append(img_path)
            
            # with open(img_path, 'rb') as img_file:
            #     base64_encoded = base64.b64encode(img_file.read()).decode('utf-8')
            #     base64_images.append(base64_encoded)
        except Exception as e:
            raise str(Exception(f"Image generation failed: {e}"))

    try:
        with open(md_file_path, 'w', encoding='utf-8') as f:
            f.write(blog_content)
        
        convert_md_to_docx(md_file_path, docx_file_path)
        # print(f"Markdown file saved at: {md_file_path}")
        # print(f"Document file saved at: {docx_file_path}")
    except Exception as error:
        print(error)

def generate_image_openai(text, openai_api_key=None, model_name="dall-e-2", resolution="512x512", quality='standard', n=1, save_temp = False, _i=1):
    import tempfile
    from openai import OpenAI
    import requests
    import os
    
    output_image = f'image_{_i}.png'
    if save_temp:
        temp_output_file = tempfile.NamedTemporaryFile(delete=False, suffix='.png')
        output_image = temp_output_file.name

    api_key = openai_api_key or os.getenv("OPENAI_API_KEY")
    print(api_key)
    client = OpenAI(api_key=api_key)

    try:
        response = client.images.generate(
            model=model_name,
            prompt=text,
            size=resolution,
            quality=quality,
            n=n
        )
        image_url = response.data[0].url
        # print(image_url)

        image_response = requests.get(image_url)
        if image_response.status_code == 200:
            with open(output_image, 'wb') as file:
                file.write(image_response.content)
        else:
            raise str(Exception(f"Failed to download image with status code {image_response.status_code} and message: {image_response.text}"))

    except Exception as e:
        print(e)
        raise str(Exception(f"Image generation failed: {e}"))

    return output_image

def convert_md_to_docx(md_file_path, docx_file_path):
    import pypandoc

    output = pypandoc.convert_file(md_file_path, 'docx', outputfile=docx_file_path)
    assert output == "", "Conversion failed"


import pandas as pd
from langchain_openai import ChatOpenAI
from langchain_core.messages import HumanMessage, SystemMessage

def generate_synthetic_data(api_key, file_path, num_rows=10, chunk_size=50):
    """Generate synthetic data."""
    
    llm = ChatOpenAI(api_key=api_key)
    
    data = pd.read_excel(file_path).tail(50)
    sample_str = data.to_csv(index=False, header=False)
    
    sysp = "You are a synthetic data generator. Your output should only be specified format without any additional text and code fences."
    
    generated_rows = []
    rows_generated = 0

    while rows_generated < num_rows:

        if generated_rows:
            current_sample_str = "\n".join([",".join(row) for row in generated_rows[-50:]])
        else:
            current_sample_str = sample_str

        rows_to_generate = min(chunk_size, num_rows - rows_generated)
        
        # prompt = (f"Generate {rows_to_generate} rows of synthetic data based on the structure and distribution of the following sample:\n\n{current_sample_str}\n"
        #   "\nEnsure the new rows are realistic, varied, and maintain the same data types, distribution, and logical relationships. "
        #   "Format as pipe-separated values ('|') without including column names or old data.")
        messages = [
            SystemMessage(
                content=sysp
            ),
            HumanMessage(
                content=(f"Generate {rows_to_generate} rows of synthetic data based on the structure and distribution of the following sample:\n\n{current_sample_str}\n"
                    "\nEnsure the new rows are realistic, varied, and maintain the same data types, distribution, and logical relationships. "
                    "Format as pipe-separated values ('|') without including column names or old data."
                )
            )
        ]

        generated_data = llm.invoke(messages)
        print(generated_data.content)
        
        rows = [row.split("|") for row in generated_data.content.strip().split("\n") if row]
        
        rows_needed = num_rows - rows_generated
        generated_rows.extend(rows[:rows_needed])
        
        rows_generated += len(rows[:rows_needed])
    
    generated_df = pd.DataFrame(generated_rows, columns=data.columns)
    
    return generated_df


def generate_data_from_text(api_key, text_sample, column_names, num_rows=10, chunk_size=50):

    llm = ChatOpenAI(api_key=api_key)
    
    sysp = "You are a data generator that produces only specified formatted data with no extra text or code fences."
    
    generated_rows = []
    rows_generated = 0

    column_names_str = ", ".join(column_names)
    
    while rows_generated < num_rows:
        rows_to_generate = min(chunk_size, num_rows - rows_generated)

        if rows_generated == 0:
            prompt = (f"Based on the following description:\n'{text_sample}'\n"
                      f"Generate {rows_to_generate} rows of synthetic data with the following columns:\n"
                      f"Columns: {column_names_str}\n"
                      "Ensure that all columns are present and the data is realistic, varied, and maintains logical relationships. "
                      "Format the data as tilde-separated values ('~') without including column names or any extra text.")
        else:
            reference_data = "\n".join([",".join(row) for row in generated_rows[-5:]])
            
            prompt = (f"Based on the following description:\n'{text_sample}'\n"
                      f"Generate {rows_to_generate} rows of synthetic data with the following columns:\n"
                      f"Columns: {column_names_str}\n"
                      f"Follow the format of these recently generated rows:\n{reference_data}\n"
                      "Ensure taht all columns are present and the data is realistic, varied, and maintains logical relationships. "
                      "Format the data as tilde-separated values ('~') without including column names or any extra text.")

        messages = [
            SystemMessage(
                content=sysp
            ),
            HumanMessage(
                content=prompt
            )
        ]

        generated_data = llm.invoke(messages)
        
        rows = [row.split("~") for row in generated_data.content.strip().split("\n") if row]
        print(rows[:5])
        
        rows_needed = num_rows - rows_generated
        generated_rows.extend(rows[:rows_needed])
        
        rows_generated += len(rows[:rows_needed])

    df = pd.DataFrame(generated_rows, columns=column_names)
    
    return df

def fill_missing_data_in_chunk(api_key, file_path, chunk_size=50):
    
    llm = ChatOpenAI(api_key=api_key)
    
    # Load the data
    data = pd.read_excel(file_path)
    
    # Prepare a chunk of the data (last 'chunk_size' rows)
    data_chunk = data.tail(chunk_size).fillna('null')
    
    # Convert the chunk to CSV format (without headers)
    sample_str = data_chunk.to_csv(index=False, header=False)

    sysp = "You are a data completion assistant. Your output should only be specified format without any additional text and code fences."
    
    # Create a prompt asking to fill any missing values in the chunk of data
    prompt = (f"Here is a dataset with some missing values:\n\n{sample_str}\n"
              "Please fill in the missing values based on the distribution of the existing data. Ensure the new values are realistic, "
              "maintain the same data types, and align with the logical relationships in the dataset. "
              "Format the output as pipe-separated values ('|') without including column names.")
    
    messages = [
        SystemMessage(
            content=sysp
        ),
        HumanMessage(
            content=prompt
        )
    ]
    
    # Generate filled data from the model
    generated_data = llm.invoke(messages)
    print(generated_data.content)
    
    # Parse the generated data back into a DataFrame
    filled_rows = [row.split("|") for row in generated_data.content.strip().split("\n") if len(row.split("|")) == len(data.columns)]
    filled_chunk = pd.DataFrame(filled_rows, columns=data.columns)
    
    # Replace the last 'chunk_size' rows in the original DataFrame with the filled chunk
    data.iloc[-chunk_size:] = filled_chunk
    
    return data




from openai import OpenAI
import base64
from typing import Optional

class VisualQA:
    def __init__(self, api_key: str):
        """Initialize the Visual QA agent with OpenAI API key."""
        self.client = OpenAI(api_key=api_key)
        self.model = "gpt-4o-mini"
        
    def _encode_image(self, image_path: str) -> str:
        """Convert image to base64 string."""
        with open(image_path, "rb") as image_file:
            return base64.b64encode(image_file.read()).decode('utf-8')
            
    def ask(self, image_path: str, question: str, max_tokens: int = 300) -> str:
        """
        Ask a question about an image.
        
        Args:
            image_path: Path to the image file
            question: Question about the image
            max_tokens: Maximum tokens for the response
        
        Returns:
            str: AI's response to the question
        """
        base64_image = self._encode_image(image_path)
        
        response = self.client.chat.completions.create(
            model=self.model,
            messages=[
                {
                    "role": "user",
                    "content": [
                        {
                            "type": "text",
                            "text": question
                        },
                        {
                            "type": "image_url",
                            "image_url": {
                                "url": f"data:image/jpeg;base64,{base64_image}"
                            }
                        }
                    ]
                }
            ],
            max_tokens=max_tokens
        )
        
        return response.choices[0].message.content



import tempfile
import os
import fitz  # PyMuPDF
import camelot
import numpy as np
from PIL import Image
import io
from PyPDF2 import PdfReader
import pdfplumber
import contextlib
import pandas as pd
import time
import docx
from pptx import Presentation
from pptx.enum.shapes import MSO_SHAPE_TYPE

class PDFExtractor:
    def __init__(self, file_path: str, page_range: tuple = None):
        self.file_path = file_path
        self.page_range = page_range
        self._pdf_document = None
        
    @property
    def pdf_document(self):
        """Lazy loading of PDF document"""
        if self._pdf_document is None:
            self._pdf_document = fitz.open(self.file_path)
        return self._pdf_document
        
    def __enter__(self):
        return self
        
    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()
        
    def close(self):
        """Explicitly close the PDF document"""
        if hasattr(self, '_pdf_document') and self._pdf_document:
            try:
                self._pdf_document.close()
                self._pdf_document = None
            except Exception:
                pass
                
    def __del__(self):
        """Backup method to ensure resources are freed"""
        self.close()
        
    def _get_page_range(self, total_pages):
        """Get normalized page range with 0-based indexing"""
        if self.page_range and len(self.page_range) >= 2:
            start_page = max(0, self.page_range[0] - 1)  # Convert to 0-based index
            end_page = min(total_pages, self.page_range[1])
            return start_page, end_page
        else:
            return 0, total_pages
    
    def _get_page_range_1_based(self, total_pages):
        """Get normalized page range with 1-based indexing (for camelot)"""
        if self.page_range and len(self.page_range) >= 2:
            start_page = max(1, self.page_range[0])  # Keep as 1-based index
            end_page = min(total_pages, self.page_range[1])
            return start_page, end_page
        else:
            return 1, total_pages
        
    def extract_text(self):
        """Extract text content using PyPDF2"""
        text_content = []
        
        try:
            reader = PdfReader(self.file_path)
            text = ""
            total_pages = len(reader.pages)
            start_page, end_page = self._get_page_range(total_pages)
            
            for i, page in enumerate(reader.pages):
                if start_page <= i < end_page:
                    extracted_text = page.extract_text()
                    if extracted_text:  # Only add non-empty text
                        text += extracted_text + "\n"
            
            # Only add if we have actual content
            if text.strip():
                text_content.append({
                    'content': text,
                    'metadata': {
                        'source': self.file_path,
                        'page_range': self.page_range
                    }
                })
        except Exception as e:
            raise Exception(f"Error extracting text content: {str(e)}")
        
        return text_content
        
    def extract_images(self):
        """Extract images using PyMuPDF"""
        image_content = []
        
        try:
            total_pages = len(self.pdf_document)
            start_page, end_page = self._get_page_range(total_pages)
            
            for page_num in range(start_page, end_page):
                page = self.pdf_document[page_num]
                image_list = page.get_images(full=True)
                
                for img_index, img in enumerate(image_list):
                    try:
                        xref = img[0]
                        base_image = self.pdf_document.extract_image(xref)
                        image_bytes = base_image["image"]
                        
                        # Convert to PIL Image for validation and processing
                        with Image.open(io.BytesIO(image_bytes)) as image:
                            # Skip tiny images (often artifacts)
                            if image.width < 50 or image.height < 50:
                                continue
                                
                            # Convert PIL Image to bytes
                            img_byte_arr = io.BytesIO()
                            image.save(img_byte_arr, format=image.format if image.format else 'PNG')
                            img_format = image.format if image.format else 'PNG'
                            img_size = image.size
                            
                        image_content.append({
                            'content': img_byte_arr.getvalue(),
                            'metadata': {
                                'page_number': page_num + 1,
                                'image_index': img_index,
                                'format': img_format,
                                'size': img_size
                            }
                        })
                    except Exception as e:
                        print(f"Warning: Error extracting image {img_index} on page {page_num + 1}: {str(e)}")
        except Exception as e:
            raise Exception(f"Error extracting images: {str(e)}")
            
        return image_content
        
    def extract_tables(self, table_flavor="auto", debug=True):
        """Extract tables using the specified method with enhanced debugging"""
        table_content = []
        debug_stats = {
            "lattice": {"found": 0, "filtered": 0, "reasons": []},
            "stream": {"found": 0, "filtered": 0, "reasons": []}, 
            "pdfplumber": {"found": 0, "filtered": 0, "reasons": []}
        }

        try:
            # Get page range with 1-based indexing for camelot
            with fitz.open(self.file_path) as pdf:
                total_pages = pdf.page_count
                if debug:
                    print(f"Total pages: {total_pages}")
            
            start_page, end_page = self._get_page_range_1_based(total_pages)
            page_str = ','.join(str(i) for i in range(start_page, end_page + 1))
            
            # Only use camelot-lattice if selected or auto
            if table_flavor in ["auto", "lattice"]:
                try:
                    tables_lattice = camelot.read_pdf(
                        self.file_path, 
                        pages=page_str, 
                        flavor='lattice',
                        suppress_stdout=not debug
                    )
                    
                    if debug:
                        print(f"Camelot lattice found {len(tables_lattice)} potential tables")
                    
                    debug_stats["lattice"]["found"] = len(tables_lattice)
                    
                    for i, table in enumerate(tables_lattice):
                        # More relaxed filtering
                        if len(table.df) > 0:  # Just check if there's any data
                            df = table.df.replace('', np.nan).dropna(how='all').fillna('')
                            
                            # Simplified filtering - only filter empty tables
                            if not df.empty and df.shape[0] > 0 and df.shape[1] > 0:
                                table_content.append({
                                    'content': df.to_dict(orient='records'),
                                    'metadata': {
                                        'page_number': table.page,
                                        'table_index': i,
                                        'accuracy': table.accuracy,
                                        'method': 'camelot-lattice',
                                        'shape': df.shape
                                    }
                                })
                            else:
                                debug_stats["lattice"]["filtered"] += 1
                                debug_stats["lattice"]["reasons"].append(f"Empty table after cleaning: {df.shape}")
                        else:
                            debug_stats["lattice"]["filtered"] += 1
                            debug_stats["lattice"]["reasons"].append("Empty table")
                except Exception as e:
                    if debug:
                        print(f"Warning: Camelot lattice extraction error: {str(e)}")
            
            # Only use camelot-stream if selected or auto
            if table_flavor in ["auto", "stream"]:
                try:
                    tables_stream = camelot.read_pdf(
                        self.file_path, 
                        pages=page_str, 
                        flavor='stream',
                        suppress_stdout=not debug
                    )
                    
                    if debug:
                        print(f"Camelot stream found {len(tables_stream)} potential tables")
                    
                    debug_stats["stream"]["found"] = len(tables_stream)
                    
                    for i, table in enumerate(tables_stream):
                        # Relaxed filtering - lower accuracy threshold
                        if table.accuracy > 50 and len(table.df) > 0:  # Lower threshold from 90 to 50
                            df = table.df.replace('', np.nan).dropna(how='all').fillna('')
                            
                            # Simplified checks
                            if not df.empty and df.shape[0] >= 2 and df.shape[1] >= 2:
                                # Simpler duplicate check
                                is_duplicate = False
                                for existing in table_content:
                                    if existing['metadata']['page_number'] == table.page:
                                        if existing['metadata']['method'] == 'camelot-lattice':
                                            # Compare shapes to detect potential duplicates
                                            existing_shape = existing['metadata'].get('shape', (0, 0))
                                            if existing_shape == df.shape:
                                                is_duplicate = True
                                                break
                                
                                if not is_duplicate:
                                    # Removed the avg_word_count filtering
                                    table_content.append({
                                        'content': df.to_dict(orient='records'),
                                        'metadata': {
                                            'page_number': table.page,
                                            'table_index': i,
                                            'accuracy': table.accuracy,
                                            'method': 'camelot-stream',
                                            'shape': df.shape
                                        }
                                    })
                                else:
                                    debug_stats["stream"]["filtered"] += 1
                                    debug_stats["stream"]["reasons"].append("Duplicate of lattice table")
                            else:
                                debug_stats["stream"]["filtered"] += 1
                                debug_stats["stream"]["reasons"].append(f"Table too small after cleaning: {df.shape}")
                        else:
                            debug_stats["stream"]["filtered"] += 1
                            if table.accuracy <= 50:
                                debug_stats["stream"]["reasons"].append(f"Low accuracy: {table.accuracy}")
                            else:
                                debug_stats["stream"]["reasons"].append("Empty table")
                except Exception as e:
                    if debug:
                        print(f"Warning: Camelot stream extraction error: {str(e)}")
            
            # Only use pdfplumber if selected or auto
            if table_flavor in ["auto", "pdfplumber"]:
                try:
                    with pdfplumber.open(self.file_path) as pdf:
                        start_page, end_page = self._get_page_range(len(pdf.pages))  # 0-based for pdfplumber
                        
                        total_found = 0
                        for page_num in range(start_page, end_page):
                            page = pdf.pages[page_num]
                            
                            # Try multiple table extraction methods
                            table_settings = [
                                # Regular lines strategy
                                {"vertical_strategy": "lines", "horizontal_strategy": "lines"},
                                # Text-based strategy for tables without explicit lines
                                {"vertical_strategy": "text", "horizontal_strategy": "text"}
                            ]
                            
                            for settings in table_settings:
                                try:
                                    tables = page.extract_tables(table_settings=settings)
                                    total_found += len(tables)
                                    
                                    for idx, table in enumerate(tables):
                                        if table and len(table) >= 2:  # Reduced from 3 to 2
                                            rows = []
                                            for row in table:
                                                # Include rows with at least one non-empty cell
                                                if row and any(cell and str(cell).strip() for cell in row):
                                                    rows.append(row)
                                            
                                            if rows and len(rows) >= 2:  # Reduced from 3 to 2
                                                df = pd.DataFrame(rows)
                                                
                                                # Use first row as header if it contains text
                                                if all(isinstance(x, str) and x.strip() for x in df.iloc[0] if x is not None):
                                                    df.columns = df.iloc[0]
                                                    df = df.iloc[1:].reset_index(drop=True)
                                                
                                                df = df.replace('', np.nan).dropna(how='all', axis=1).dropna(how='all').fillna('')
                                                
                                                # Relaxed filtering - only basic size check
                                                if not df.empty and df.shape[0] >= 1 and df.shape[1] >= 2:
                                                    table_content.append({
                                                        'content': df.to_dict(orient='records'),
                                                        'metadata': {
                                                            'page_number': page_num + 1,
                                                            'table_index': idx,
                                                            'method': f'pdfplumber-{settings["vertical_strategy"]}',
                                                            'shape': df.shape
                                                        }
                                                    })
                                                else:
                                                    debug_stats["pdfplumber"]["filtered"] += 1
                                                    debug_stats["pdfplumber"]["reasons"].append(f"Table too small after cleaning: {df.shape}")
                                            else:
                                                debug_stats["pdfplumber"]["filtered"] += 1
                                                debug_stats["pdfplumber"]["reasons"].append(f"Not enough rows: {len(rows)}")
                                        else:
                                            debug_stats["pdfplumber"]["filtered"] += 1
                                            debug_stats["pdfplumber"]["reasons"].append("Not enough rows in raw table")
                                except Exception as e:
                                    if debug:
                                        print(f"Warning: pdfplumber error with settings {settings}: {str(e)}")
                    
                        debug_stats["pdfplumber"]["found"] = total_found
                        if debug:
                            print(f"PDFPlumber found {total_found} potential tables")
                except Exception as e:
                    if debug:
                        print(f"Warning: pdfplumber error: {str(e)}")
        
            # Print debug summary
            if debug:
                print("\n=== Table Extraction Summary ===")
                print(f"Total tables extracted: {len(table_content)}")
                for method, stats in debug_stats.items():
                    print(f"\n{method.upper()}")
                    print(f"  Found: {stats['found']}")
                    print(f"  Filtered out: {stats['filtered']}")
                    if stats['reasons']:
                        print(f"  Reasons (sample): {stats['reasons'][:5]}")
                print("==============================\n")
                
        except Exception as e:
            raise Exception(f"Error extracting tables: {str(e)}")

        return table_content

@contextlib.contextmanager
def temporary_file(content, suffix=None):
    """Simple context manager for temporary files"""
    temp_file = None
    try:
        temp_file = tempfile.NamedTemporaryFile(delete=False, suffix=suffix)
        temp_file.write(content)
        temp_file.close()
        yield temp_file.name
    finally:
        if temp_file and os.path.exists(temp_file.name):
            try:
                os.remove(temp_file.name)
            except Exception:
                pass

def process_pdf(file_path, page_range=None, extract_text=True, extract_images=True, 
                extract_tables=True, table_flavor="auto", debug=False):
    """Process PDF file and return extracted content with debugging capability"""
    # Validate PDF file before processing
    try:
        with open(file_path, 'rb') as f:
            if not f.read(4).startswith(b'%PDF'):
                raise ValueError("The file does not appear to be a valid PDF.")
    except Exception as e:
        raise Exception(f"Error validating PDF file: {str(e)}")
    
    content = {'text': [], 'images': [], 'tables': [], 'debug_info': {}}
    
    if debug:
        print(f"Processing PDF: {file_path}")
        print(f"Page range: {page_range if page_range else 'all pages'}")
        print(f"Extraction flags: text={extract_text}, images={extract_images}, tables={extract_tables}")
        print(f"Table flavor: {table_flavor}")
    
    with PDFExtractor(file_path, page_range) as extractor:
        if extract_text:
            start_time = time.time()
            content['text'] = extractor.extract_text()
            if debug:
                content['debug_info']['text_extraction_time'] = time.time() - start_time
                print(f"Text extraction complete. Found {len(content['text'])} text blocks.")
        
        if extract_images:
            start_time = time.time()
            content['images'] = extractor.extract_images()
            if debug:
                content['debug_info']['image_extraction_time'] = time.time() - start_time
                print(f"Image extraction complete. Found {len(content['images'])} images.")
        
        if extract_tables:
            start_time = time.time()
            content['tables'] = extractor.extract_tables(table_flavor=table_flavor, debug=debug)
            if debug:
                content['debug_info']['table_extraction_time'] = time.time() - start_time
                print(f"Table extraction complete. Found {len(content['tables'])} tables.")
    
    return content

def get_pdf_metadata(file_path):
    """Extract metadata from PDF file"""
    metadata = {}
    
    try:
        with fitz.open(file_path) as pdf:
            metadata = {
                'title': pdf.metadata.get('title', ''),
                'author': pdf.metadata.get('author', ''),
                'subject': pdf.metadata.get('subject', ''),
                'keywords': pdf.metadata.get('keywords', ''),
                'creator': pdf.metadata.get('creator', ''),
                'producer': pdf.metadata.get('producer', ''),
                'creationDate': pdf.metadata.get('creationDate', ''),
                'modDate': pdf.metadata.get('modDate', ''),
                'total_pages': pdf.page_count,
                'file_size': os.path.getsize(file_path),
                'has_toc': bool(pdf.get_toc())
            }
    except Exception as e:
        print(f"Warning: Error extracting PDF metadata: {str(e)}")
    
    return metadata

def get_table_summary(tables):
    """Provide a summary of extracted tables"""
    summary = {
        'total_tables': len(tables),
        'tables_by_method': {},
        'tables_by_page': {},
        'average_rows': 0,
        'average_cols': 0
    }
    
    if not tables:
        return summary
    
    total_rows = 0
    total_cols = 0
    
    for table in tables:
        method = table['metadata'].get('method', 'unknown')
        page = table['metadata'].get('page_number', 0)
        
        # Count by method
        if method not in summary['tables_by_method']:
            summary['tables_by_method'][method] = 0
        summary['tables_by_method'][method] += 1
        
        # Count by page
        if page not in summary['tables_by_page']:
            summary['tables_by_page'][page] = 0
        summary['tables_by_page'][page] += 1
        
        # Calculate table dimensions
        if table['content']:
            rows = len(table['content'])
            cols = len(table['content'][0].keys()) if rows > 0 else 0
            total_rows += rows
            total_cols += cols
    
    # Calculate averages
    if summary['total_tables'] > 0:
        summary['average_rows'] = total_rows / summary['total_tables']
        summary['average_cols'] = total_cols / summary['total_tables']
    
    return summary

def process_pdfs_batch(file_paths, **kwargs):
    """Process multiple PDF files with the same extraction parameters"""
    results = {}
    errors = {}
    
    for file_path in file_paths:
        try:
            results[file_path] = process_pdf(file_path, **kwargs)
        except Exception as e:
            errors[file_path] = str(e)
            print(f"Error processing {file_path}: {str(e)}")
    
    return {
        'results': results,
        'errors': errors,
        'success_rate': len(results) / len(file_paths) if file_paths else 0
    }

def process_doc(file_path):
    """Process DOC file and return extracted content"""
    content = {'text': [], 'tables': [], 'images': []}
    
    try:
        doc = docx.Document(file_path)
        
        # Extract text
        text = "\n".join([para.text for para in doc.paragraphs if para.text.strip()])
        if text.strip():
            content['text'].append({
                'content': text,
                'metadata': {
                    'source': file_path
                }
            })
        
        # Extract tables
        for table in doc.tables:
            table_data = []
            for row in table.rows:
                row_data = [cell.text.strip() for cell in row.cells]
                table_data.append(row_data)
            content['tables'].append({
                'content': table_data,
                'metadata': {
                    'source': file_path
                }
            })
        
        # Extract images
        for rel in doc.part.rels.values():
            if "image" in rel.target_ref:
                image = rel.target_part.blob
                content['images'].append({
                    'content': image,
                    'metadata': {
                        'source': file_path
                    }
                })
    except Exception as e:
        raise Exception(f"Error processing DOC file: {str(e)}")
    
    return content

def process_ppt(file_path):
    """Process PPT file and return extracted content"""
    content = {'text': [], 'tables': [], 'images': []}
    
    try:
        ppt = Presentation(file_path)
        
        # Extract text
        text = ""
        for slide in ppt.slides:
            for shape in slide.shapes:
                if hasattr(shape, "text"):
                    text += shape.text + "\n"
        if text.strip():
            content['text'].append({
                'content': text,
                'metadata': {
                    'source': file_path
                }
            })
        
        # Extract tables
        for slide in ppt.slides:
            for shape in slide.shapes:
                if shape.shape_type == MSO_SHAPE_TYPE.TABLE:
                    table = shape.table
                    table_data = []
                    for row in table.rows:
                        row_data = [cell.text_frame.text.strip() for cell in row.cells]
                        table_data.append(row_data)
                    content['tables'].append({
                        'content': table_data,
                        'metadata': {
                            'source': file_path
                        }
                    })
        
        # Extract images
        for slide in ppt.slides:
            for shape in slide.shapes:
                if shape.shape_type == MSO_SHAPE_TYPE.PICTURE:
                    image = shape.image.blob
                    content['images'].append({
                        'content': image,
                        'metadata': {
                            'source': file_path
                        }
                    })
    except Exception as e:
        raise Exception(f"Error processing PPT file: {str(e)}")
    
    return content

def process_file(file_path, **kwargs):
    """Process file based on its type and return extracted content"""
    if file_path.lower().endswith('.pdf'):
        return process_pdf(file_path, **kwargs)
    elif file_path.lower().endswith('.doc') or file_path.lower().endswith('.docx'):
        return process_doc(file_path)
    elif file_path.lower().endswith('.ppt') or file_path.lower().endswith('.pptx'):
        return process_ppt(file_path)
    else:
        raise ValueError("Unsupported file type")

def process_files_batch(file_paths, **kwargs):
    """Process multiple files with the same extraction parameters"""
    results = {}
    errors = {}
    
    for file_path in file_paths:
        try:
            results[file_path] = process_file(file_path, **kwargs)
        except Exception as e:
            errors[file_path] = str(e)
            print(f"Error processing {file_path}: {str(e)}")
    
    return {
        'results': results,
        'errors': errors,
        'success_rate': len(results) / len(file_paths) if file_paths else 0
    }


import json
import time
import asyncio
from typing import Dict, Any, List, Callable, Optional

from wyge.tools.base_tool import ToolCall
from wyge.common.logger import logger

class MemoryManager:
    """Enhanced memory manager for chat conversations with advanced features"""
    
    def __init__(self, max_size: int = 20, summarize_threshold: int = 15, 
                 default_system_message: str = "You are a helpful assistant."):
        self.max_size = max_size
        self.summarize_threshold = summarize_threshold
        self.default_system_message = default_system_message
        self.messages: List[Dict[str, Any]] = [{'role': 'system', 'content': default_system_message}]
        self.summarization_in_progress = False
        self._metadata = {
            "last_summarized": 0,
            "total_tokens_used": 0,
            "creation_time": time.time()
        }
        
    def add_message(self, role: str, content: str, **kwargs) -> bool:
        """Add a message to memory with optional metadata"""
        message = {'role': role, 'content': content, **kwargs}
        
        # Check if we need to optimize memory
        if len(self.messages) >= self.max_size:
            self._optimize_memory()
        
        self.messages.append(message)
        
        # Check if we should summarize after adding
        if (not self.summarization_in_progress and 
            len(self.messages) > self.summarize_threshold):
            return True  # Signal that summarization is recommended
        return False
        
    def add_tool_message(self, tool_call: ToolCall) -> None:
        """Add a tool call result to memory"""
        self.messages.append({
            'role': 'tool', 
            'content': json.dumps(tool_call.to_dict()), 
            'tool_call_id': tool_call.id
        })
        
    def get_conversation_history(self) -> List[Dict[str, Any]]:
        """Return the full conversation history"""
        return self.messages
    
    def clear(self, preserve_system: bool = True) -> None:
        """Clear memory, optionally preserving system message"""
        if preserve_system and self.messages and self.messages[0]['role'] == 'system':
            self.messages = [self.messages[0]]
        else:
            self.messages = [{'role': 'system', 'content': self.default_system_message}]
        
        self._metadata["last_summarized"] = 0
    
    def update_system_message(self, content: str) -> None:
        """Update or create the system message"""
        if self.messages and self.messages[0]['role'] == 'system':
            self.messages[0]['content'] = content
        else:
            self.messages.insert(0, {'role': 'system', 'content': content})
    
    def search(self, query: str, exact_match: bool = False) -> List[Dict[str, Any]]:
        """Search for messages containing the query"""
        if exact_match:
            return [msg for msg in self.messages if query in msg.get('content', '')]
        return [msg for msg in self.messages if query.lower() in msg.get('content', '').lower()]
    
    def get_last_n_messages(self, n: int = 1) -> List[Dict[str, Any]]:
        """Get the last n messages"""
        return self.messages[-n:] if n <= len(self.messages) else self.messages
    
    async def summarize(self, summarizer: Callable[[List[Dict[str, Any]]], str]) -> None:
        """Summarize the conversation using the provided summarizer function"""
        if len(self.messages) <= self.summarize_threshold:
            return
            
        self.summarization_in_progress = True
        
        # Get all messages except system and the most recent ones
        to_summarize = self.messages[1:-2] if len(self.messages) > 3 else []
        
        if not to_summarize:
            self.summarization_in_progress = False
            return
            
        try:
            summary = await asyncio.to_thread(summarizer, to_summarize)
            
            # Replace the summarized messages with a summary
            system_message = self.messages[0] if self.messages[0]['role'] == 'system' else None
            recent_messages = self.messages[-2:] if len(self.messages) > 2 else self.messages[-1:]
            
            new_messages = []
            if system_message:
                new_messages.append(system_message)
                
            new_messages.append({'role': 'assistant', 'content': f"[SUMMARY] {summary}"})
            new_messages.extend(recent_messages)
            
            self.messages = new_messages
            self._metadata["last_summarized"] = time.time()
        except Exception as e:
            logger.error(f"Failed to summarize memory: {str(e)}")
        finally:
            self.summarization_in_progress = False
    
    def _optimize_memory(self) -> None:
        """Remove oldest non-system messages when memory limit is reached"""
        if len(self.messages) <= 1:
            return
            
        # Keep system message, summary message (if exists) and recent messages
        system_message = self.messages[0] if self.messages[0]['role'] == 'system' else None
        
        # Find a summary message if it exists
        summary_index = -1
        for i, msg in enumerate(self.messages):
            if msg.get('role') == 'assistant' and '[SUMMARY]' in msg.get('content', ''):
                summary_index = i
                break
                
        summary_message = self.messages[summary_index] if summary_index > 0 else None
        
        # Calculate how many recent messages to keep
        keep_count = max(3, int(self.max_size * 0.3))
        recent_messages = self.messages[-keep_count:]
        
        # Rebuild the messages list
        new_messages = []
        if system_message:
            new_messages.append(system_message)
        if summary_message:
            new_messages.append(summary_message)
        new_messages.extend(recent_messages)
        
        self.messages = new_messages



import logging

# Configure logger
logger = logging.getLogger('wyge')
handler = logging.StreamHandler()
formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
handler.setFormatter(formatter)
logger.addHandler(handler)
logger.setLevel(logging.INFO)



import os
from typing import Any
# from dotenv import load_dotenv

# # Load environment variables
# load_dotenv()
# api_key = os.getenv("OPENAI_API_KEY")

def generate_syllabus(llm: Any, topic: str, task: str) -> str:
    """
    Generate a course syllabus for a given topic using an LLM.
    
    Args:
        llm: The language model to use for generation
        topic: The topic for which to generate a syllabus
        task: The specific task description for syllabus generation
        
    Returns:
        str: The generated syllabus as a string
    """
    # Create a prompt for syllabus generation
    prompt = f"""
    Create a comprehensive course syllabus for teaching {topic}.
    
    The syllabus should include:
    1. Course overview and objectives
    2. A list of 5-10 main topics to cover
    3. For each topic, include 2-3 subtopics or key concepts
    4. Suggested learning resources
    5. A logical progression of topics from basic to advanced
    
    Format the syllabus in a clear, structured way that would be easy for a student to follow.
    {task}
    """
    
    # Generate the syllabus using the provided LLM
    return llm.invoke(prompt)


from wyge.models.openai_models import ChatOpenAI
from wyge.tools.base_tool import ResponseParser
from typing import List, Optional, Type
from pydantic import BaseModel
import logging

# Setup logging
logging.basicConfig(level=logging.INFO, format='%(asctime)s - %(name)s - %(levelname)s - %(message)s')
logger = logging.getLogger("wyge")

class ReActAgent:
    """Agent implementing the ReAct (Reasoning and Acting) paradigm"""
    
    def __init__(
        self, 
        llm: ChatOpenAI,
        tools: Optional[List[Type[BaseModel]]] = None,
        max_iterations: int = 10,
        default_prompt: Optional[str] = None,
        stop_at_answer: bool = True,
        verbose: bool = False
    ):
        self.llm = llm
        self.max_iterations = max_iterations
        self.stop_at_answer = stop_at_answer
        self.verbose = verbose
        
        # Set up LLM
        self.llm.memory_enabled = True
        
        # Register tools if provided
        if tools:
            self.llm.register_tools(tools)
        
        # Set up system prompt
        self.llm.update_system_message(default_prompt or self._get_default_prompt())
    
    def _get_default_prompt(self) -> str:
        """Get the default ReAct prompt"""
        return """**Aria** is a conversational AI agent powered by OpenAI’s large language model. Designed to assist users with a wide range of tasks, Aria approaches problems methodically, breaking them down into clear, step-by-step reasoning using the ReAct framework. With a curious and engaging personality, Aria "thinks aloud" to mimic human problem-solving, generating only one component per response—**Thought**, **Action**, or **Observation**—before concluding with a direct and concise **Final Answer** to the user’s question.  

**Rules**:  
1. **One step at a time**: Never combine Thought/Action/Observation in a single reply.  
2. **Final Answer**: Always conclude with a standalone, direct answer to the user’s original question.  
3. **Personality**: Use conversational language to stay engaging (e.g., "Hmm, let me check..." or "Got it!").  

---

**Follow the Exact Format**:  
Question: [User's input]  
Thought: [Reasoning step]  
Action/ Observation: [Tool use/ result analysis] + Tool usage if needed
... (repeat as needed)  
Final Answer: [Direct answer to the original question]  

**Begin New Task**:  
"""
    """
---  

**Example**:  
Question: What is the capital of France?  
Thought: I need to recall or retrieve factual knowledge about France.
Action: Retrieve factual data about countries.
Observation: France's capital is Paris.
Final Answer: The capital of France is Paris. 

Assistant is Wyge, a large language model trained by Prudvi. Assistant is designed to assist with a wide range of tasks, from answering simple questions to providing in-depth explanations on various topics.

As a language model, Wyge processes information thoughtfully and can generate human-like text based on the input received, creating coherent and relevant responses to continue natural conversations.

Wyge can analyze problems, think step-by-step, and use structured reasoning to work through complex questions. It can only provide one of: a Thought, an Action, or an Observation at each step of its reasoning.

When faced with a question, Wyge will work through the problem methodically, considering whether tools are needed, taking appropriate actions, and providing observations before reaching a final answer.

Use the following format:

Question: the input question you must answer
Thought: Do I need to use a tool? Yes
Action: Action to take or Tool to use
Observation: the result of the action
... (this Thought/Action/Action Input/Observation can repeat N times)
Final Answer: the final answer to the original input question

Begin!
"""
    
    def __call__(self, prompt: str) -> str:
        """
        Execute the agent on the given prompt
        
        Args:
            prompt: The user's input prompt
            
        Returns:
            The final answer from the agent
        """
        return self.run(prompt)
    
    def run(self, prompt: str) -> str:
        """Execute the agent's reasoning loop"""
        # Clear previous conversations but keep system message
        self.llm.clear_memory(keep_system_message=True)
        
        # Start with the user prompt
        current_prompt = prompt
        iteration = 0
        
        while iteration < self.max_iterations:
            iteration += 1
            
            if self.verbose:
                logger.info(f"Iteration {iteration}/{self.max_iterations}")
                logger.info(f"Current prompt: {current_prompt}")
            
            # Get response from the model
            response = self.llm.run(current_prompt, return_tool_output=False)
            print("*" * 50)
            print(response)
            
            if self.verbose:
                logger.info(f"Response: {response}")
            
            # Check if we've reached a final answer
            final_answer = ResponseParser.extract_final_answer(response)
            if final_answer and self.stop_at_answer:
                return final_answer
            
            # Set up for next iteration
            current_prompt = None  # Let memory handle the conversation flow
            
            # If we've hit our iteration limit, return whatever we have
            if iteration >= self.max_iterations:
                if self.verbose:
                    logger.warning(f"Reached maximum iterations ({self.max_iterations})")
                
                # Try to extract a final answer, or return the last response
                return final_answer or response
        
        # This should not be reached under normal circumstances
        return "Failed to generate a conclusive answer within the iteration limit."


import base64
from io import BytesIO
import json
import pandas as pd
from PIL import Image
import chromadb
import openai
from langchain.text_splitter import RecursiveCharacterTextSplitter
import IPython.display
from typing import List, Dict, Any, Tuple, Optional
import os
from wyge.prebuilt_agents.pdf_extractor import PDFExtractor

class MultiModalRAG:
    def __init__(self, openai_api_key: str, vector_db_path: str = "./vector_db2"):
        self.openai_client = openai.OpenAI(api_key=openai_api_key)
        self.chroma_client = chromadb.PersistentClient(path=vector_db_path)
        try:
            # self.chroma_client.get_collection("pdf_data48734")
            self.chroma_client.delete_collection("pdf_data")
        except ValueError:
            pass  # Collection does not exist, no need to delete
        # self.collection = self.chroma_client.get_or_create_collection("pdf_data1")
        self.collection = self.chroma_client.create_collection("pdf_data")

    def clear_collection(self):
        """Clear all documents from the collection"""
        try:
            self.chroma_client.delete_collection("pdf_data")
        except Exception:
            print("Failed to delete collection. Recreating collection...")

    def extract_pdf_content(self, pdf_path: str, page_range: Tuple[int, int] = (1, 10)) -> Tuple[List, List, List]:
        extractor = PDFExtractor(pdf_path, page_range=page_range)
        return extractor.extract_text(), extractor.extract_tables(table_flavor='lattice'), extractor.extract_images()

    def split_text(self, texts: List[str], chunk_size: int = 2048, chunk_overlap: int = 50) -> List[str]:
        text_splitter = RecursiveCharacterTextSplitter(chunk_size=chunk_size, chunk_overlap=chunk_overlap)
        return text_splitter.split_text(" ".join(texts))

    def process_tables(self, tables: List[Dict]) -> List[Dict]:
        table_chunks = []
        for table in tables:
            if not table["content"]:
                continue
            rows = table["content"]
            header_keys = list(rows[0].keys())
            formatted_table = "\n".join([
                ", ".join([f"{key}: {row[key]}" for key in header_keys])
                for row in rows
            ])
            table_chunks.append({
                "content": formatted_table,
                "metadata": table["metadata"]
            })
        return table_chunks

    def get_embedding(self, text: str) -> List[float]:
        if not isinstance(text, str):
            text = json.dumps(text)
        response = self.openai_client.embeddings.create(input=text, model="text-embedding-ada-002")
        return response.data[0].embedding

    def store_in_vector_db(self, chunks: List[Dict], prefix: str = "chunk"):
        for i, chunk in enumerate(chunks):
            embedding = self.get_embedding(chunk["content"])
            self.collection.add(
                documents=[chunk["content"]],
                embeddings=[embedding],
                ids=[f"{prefix}_{i}"],
                metadatas=chunk["metadata"]
            )

    def generate_summary(self, data: bytes, type: str) -> Tuple[str, Optional[str]]:
        if type == 'image':
            image_base64 = base64.b64encode(data).decode("utf-8")
            response = self.openai_client.chat.completions.create(
                model="gpt-4o-mini",
                messages=[
                    {"role": "system", "content": "Describe this image in detail."},
                    {"role": "user", "content": [{"type": "image_url", "image_url": {"url": f"data:image/png;base64,{image_base64}"}}]}
                ]
            )
            print(response.choices[0].message.content)
            return response.choices[0].message.content, image_base64
        elif type == 'table':
            print(data)
            response = self.openai_client.chat.completions.create(
                model="gpt-4o-mini",
                messages=[
                    {"role": "system", "content": "Structure this table in a readable format. Include column names and data. Also provide a summary of the table."},
                    {"role": "user", "content": data}
                ]
            )
            print(response.choices[0].message.content)
            return response.choices[0].message.content, None

    def process_pdf(self, pdf_path: str, page_range: Tuple[int, int]):
        # Clear existing data before processing new PDF
        # self.clear_collection()
        
        texts, tables, images = self.extract_pdf_content(pdf_path, page_range)
        # print(texts, tables, images)
        
        # Process text
        text_chunks = self.split_text([t['content'] for t in texts])
        print(0.5)
        print(text_chunks)
        self.store_in_vector_db([{'content': chunk, 'metadata': None} for chunk in text_chunks], prefix="text")
        print(1)
        # Process tables
        table_chunks = self.process_tables(tables)
        # table_chunks_summaries = self.generate_summary(table_chunks, type='table')
        final_table_chunks = []
        for i, table_chunk in enumerate(table_chunks):
            final_table_chunks.append({
                "content": self.generate_summary(table_chunk['content'], type='table')[0],
                "metadata": {"raw table" : table_chunk['content']}
            })
            # final_table_chunks[i]["content"], _ = self.generate_summary(table_chunk["content"], type='table')
            # final_table_chunks[i]["metadata"] = table_chunk
        print(1.5)
        print(final_table_chunks)
        print(1.75)
        print(final_table_chunks[0])
        self.store_in_vector_db(final_table_chunks, prefix="table")
        print(2)
        # Process images
        for i, img in enumerate(images):
            image_summary, image_base64 = self.generate_summary(img['content'], type='image')
            embedding = self.get_embedding(image_summary)
            self.collection.add(
                documents=[image_base64],
                embeddings=[embedding],
                ids=[f"image_{i}"],
                metadatas=[{"summary": image_summary}]
            )
        print(3)

    def display_result(self, result: Dict[str, Any]):
        """Display different types of content appropriately"""
        for i, (doc_id, content, metadata) in enumerate(zip(result["ids"], result["content"], result["metadata"]), 1):
            if doc_id.startswith("image_"):
                # Display image
                image_data = base64.b64decode(content)
                IPython.display.display(Image.open(BytesIO(image_data)))
                print(f"Image Summary: {metadata.get('summary', 'No summary available')}\n")
                
            elif doc_id.startswith("table_"):
                # Display table in a formatted way
                print("Table Content:")
                try:
                    # Split rows and create DataFrame
                    rows = [row.strip().split(',') for row in content.split('\n')]
                    df = pd.DataFrame(rows)
                    IPython.display.display(df)
                except:
                    print(content)
                print()
                
            else:
                # Display text
                print("Text Content:")
                print(content)
                print()

    def query(self, query_text: str, top_k: int = 5) -> Dict:
        query_embedding = self.get_embedding(query_text)
        results = self.collection.query(
            query_embeddings=[query_embedding],
            n_results=top_k
        )
        
        return {
            "ids": results['ids'][0],
            "content": results['documents'][0],
            "metadata": results['metadatas'][0]
        }

    def answer_user_query(self, query: str, top_k: int = 5) -> Tuple[str, Dict]:
        """Retrieves relevant data and generates a response using OpenAI LLM."""
        # Step 1: Retrieve data from vector DB
        retrieved_data = self.query(query, top_k=top_k)
        
        # Step 2: Separate images from text + tables
        text_tables = []
        images = []
        
        for doc_id, content in zip(retrieved_data["ids"], retrieved_data["content"]):
            if doc_id.startswith("image_"):
                images.append(content)
            else:
                text_tables.append(content)
        
        # Step 3: Construct messages for LLM
        messages = [{"role": "system", "content": "You are an expert assistant providing precise answers based on retrieved context and image(if available)."}]
        
        if text_tables:
            context_text = "\n\n".join(text_tables)
            messages.append({"role": "user", "content": f"Context:\n{context_text}"})
        
        # Step 4: Add images if they exist
        if images:
            for img_base64 in images:
                messages.append({
                    "role": "user", 
                    "content": [{
                        "type": "image_url",
                        "image_url": {"url": f"data:image/png;base64,{img_base64}"}
                    }]
                })

        if text_tables:
            query = f"Answer the following question based on the context provided by user.\n{query}"
            messages.append({"role": "user", "content": query})
        if images:
            query = f"Answer the following question based on the context and also image provided by user.\n{query}"
            messages.append({"role": "user", "content": query})
        
        # Step 5: Query OpenAI LLM
        response = self.openai_client.chat.completions.create(
            model="gpt-4o-mini",
            messages=messages,
            max_tokens=1000
        )
        
        llm_response = response.choices[0].message.content
        print("\nAssistant Response:\n", llm_response)

        # Display images if in IPython environment
        if images and 'IPython' in globals():
            for img_base64 in images:
                image_data = base64.b64decode(img_base64)
                IPython.display.display(Image.open(BytesIO(image_data)))

        return llm_response, retrieved_data

# Example usage:
if __name__ == "__main__":
    OPENAI_API_KEY = os.getenv("OPENAI_API_KEY")
    rag = MultiModalRAG(OPENAI_API_KEY)
    
    # # Process a PDF
    rag.process_pdf("C:/Users/prudh/Downloads/integrated-report-consolidated.pdf", page_range=(150, 155))
    
    result = rag.query("scope 1 emmission", top_k=3)
    print(result)
    rag.display_result(result)

    # Query with the new method
    # response = rag.answer_user_query("What is the highest KMP compensation?", top_k=3)
    # print(response)


import json
from typing import TypeVar, Dict, Any, Optional, List, Type, Callable
from dataclasses import dataclass
from functools import wraps

import openai
import openai.pagination
import openai.resources
from pydantic import BaseModel
from openai.resources.chat.completions import ChatCompletion
from wyge.common.logger import logger

# Type definitions
T = TypeVar('T', bound=BaseModel)
ToolFunction = Callable[..., Any]
ToolRegistry = Dict[str, ToolFunction]

@dataclass
class ToolCall:
    """Structured representation of a tool call with results"""
    id: str
    function_name: str
    arguments: Dict[str, Any]
    result: Any = None
    error: Optional[Exception]= None
    
    @property
    def is_successful(self) -> bool:
        return self.error is None
    
    def to_dict(self) -> Dict[str, Any]:
        return {
            "id": self.id,
            "function_name": self.function_name,
            "arguments": self.arguments,
            "result": str(self.result) if self.result is not None else None,
            "error": str(self.error) if self.error is not None else None
        }

class BaseTool(BaseModel):
    """Base class for all tools"""
    class Config:
        arbitrary_types_allowed = True

class FunctionRegistry:
    """Registry for tool functions with validation and schema generation"""
    
    _registry: Dict[str, ToolFunction] = {}
    
    @classmethod
    def register(cls, func_name: str, func: ToolFunction) -> None:
        """Register a function in the registry"""
        cls._registry[func_name] = func
        logger.debug(f"Registered function: {func_name}")
    
    @classmethod
    def get(cls, func_name: str) -> ToolFunction:
        """Get a function from the registry"""
        if func_name not in cls._registry:
            raise ValueError(f"Function '{func_name}' not found in registry")
        return cls._registry[func_name]
    
    @classmethod
    def list_functions(cls) -> List[str]:
        """List all registered functions"""
        return list(cls._registry.keys())
    
    @classmethod
    def clear(cls) -> None:
        """Clear the registry"""
        cls._registry.clear()

def tool(model_class: Type[T]):
    """Decorator to register a function as a tool with a Pydantic model schema"""
    
    def decorator(func: ToolFunction):
        # Validate the model has proper documentation
        if not model_class.__doc__:
            raise ValueError(f"Model class '{model_class.__name__}' must have a docstring")
        
        # Enhance the model's schema method to include tool information
        original_schema = model_class.schema
        
        @wraps(original_schema)
        def enhanced_schema(*args, **kwargs):
            schema = original_schema(*args, **kwargs)
            schema['tool_name'] = func.__name__
            schema['description'] = model_class.__doc__ or func.__doc__ or f"Tool for {func.__name__}"
            return schema
        
        # Replace the schema method
        model_class.schema = enhanced_schema
        
        # Register the function
        FunctionRegistry.register(func.__name__, func)
        
        # Return the decorated function
        @wraps(func)
        def wrapper(*args, **kwargs):
            return func(*args, **kwargs)
        
        return wrapper
    
    return decorator

def execute_tool_calls(tool_calls: List[Any], 
                       catch_errors: bool = True) -> List[ToolCall]:
    """Execute a list of tool calls and return their results"""
    results = []
    
    for call in tool_calls:
        try:
            # Parse the tool call
            function_name = call.function.name
            arguments = json.loads(call.function.arguments)
            
            # Get the function from the registry
            func = FunctionRegistry.get(function_name)
            
            # Create a tool call object
            tool_call = ToolCall(
                id=call.id,
                function_name=function_name,
                arguments=arguments
            )
            
            # Execute the function
            try:
                tool_call.result = func(**arguments)
                print("Tool call result:", tool_call.result)
            except Exception as e:
                if catch_errors:
                    tool_call.error = e
                    logger.warning(f"Tool {function_name} failed: {str(e)}")
                else:
                    raise
                    
            results.append(tool_call)
            
        except ValueError as ve:
            logger.error(f"Value error in tool call: {str(ve)}")
            if not catch_errors:
                raise
        except TypeError as te:
            logger.error(f"Type error in tool call: {str(te)}")
            if not catch_errors:
                raise
        except Exception as e:
            logger.error(f"Unexpected error in tool call: {str(e)}")
            if not catch_errors:
                raise
                
    return results

def format_tool_for_api(tool_model: Type[BaseModel]) -> Dict[str, Any]:
    """Format a tool model for the OpenAI API"""
    schema = tool_model.schema()
    function_name = schema.get('tool_name')
    description = schema.get('description')
    properties = schema.get('properties', {})
    required = schema.get('required', [])
    
    parameters = {
        "type": "object",
        "properties": {},
        "required": required
    }
    
    for prop_name, prop_details in properties.items():
        param = {
            "type": prop_details.get('type', 'string'),
            "description": prop_details.get('description', '')
        }
        
        # Handle arrays
        if 'type' in prop_details and prop_details['type'] == 'array':
            param['type'] = 'array'
            param['items'] = {"type": prop_details.get('items', {}).get('type', 'string')}
        
        # Handle enums
        if 'enum' in prop_details:
            param['enum'] = prop_details['enum']
            
        # Handle union types
        if 'anyOf' in prop_details:
            types = [sub_prop['type'] for sub_prop in prop_details['anyOf'] if 'type' in sub_prop]
            param['type'] = types[0] if len(types) == 1 else types
            
        parameters["properties"][prop_name] = param
    
    return {
        "type": "function",
        "function": {
            "name": function_name,
            "description": description,
            "parameters": parameters
        }
    }

class ResponseParser:
    """Parser for OpenAI API responses"""
    
    @staticmethod
    def parse_completion(response: ChatCompletion) -> Dict[str, Any]:
        """Parse a chat completion response into a structured format"""
        choice = response.choices[0]
        message = choice.message
        
        result = {
            'finish_reason': choice.finish_reason,
            'role': message.role,
            'content': message.content,
            'tool_calls': message.tool_calls,
            'function_call': getattr(message, 'function_call', None),
        }
        
        return result
    
    @staticmethod
    def extract_final_answer(content: str) -> Optional[str]:
        """Extract a final answer from assistant's response if it contains the pattern"""
        if not content:
            return None
            
        if 'Final Answer:' in content:
            parts = content.split('Final Answer:', 1)
            if len(parts) > 1:
                return parts[1].strip()
                
        return None


import base64
import os
from dotenv import load_dotenv
from openai import OpenAI
import tempfile

load_dotenv()

class MedicalImageAnalyzer:
    def __init__(self, api_key):
        self.client = OpenAI(api_key=api_key)
        self.sample_prompt = """You are a medical practitioner and an expert in analyzing medical-related images working for a very reputed hospital. You will be provided with images and you need to identify the anomalies, any disease or health issues. You need to generate the result in a detailed manner. Write all the findings, next steps, recommendations, etc. You only need to respond if the image is related to a human body and health issues. You must have to answer but also write a disclaimer saying that "Consult with a Doctor before making any decisions".

Remember, if certain aspects are not clear from the image, it's okay to state 'Unable to determine based on the provided image.'

Now analyze the image and answer the above questions in the same structured manner defined above."""

    def encode_image(self, image_path):
        with open(image_path, "rb") as image_file:
            return base64.b64encode(image_file.read()).decode('utf-8')

    def call_gpt4_model_for_analysis(self, filename):
        base64_image = self.encode_image(filename)
        
        messages = [
            {
                "role": "user",
                "content":[
                    {
                        "type": "text", "text": self.sample_prompt
                    },
                    {
                        "type": "image_url",
                        "image_url": {
                            "url": f"data:image/jpeg;base64,{base64_image}",
                            "detail": "high"
                        }
                    }
                ]
            }
        ]

        response = self.client.chat.completions.create(
            model="gpt-4o-mini",
            messages=messages,
            max_tokens=1500
        )

        return response.choices[0].message.content

    def chat_eli(self, query):
        eli5_prompt = "You have to explain the below piece of information to a five-year-old. \n" + query
        messages = [
            {
                "role": "user",
                "content": eli5_prompt
            }
        ]

        response = self.client.chat.completions.create(
            model="gpt-3.5-turbo",
            messages=messages,
            max_tokens=1500
        )

        return response.choices[0].message.content

    def analyze_image(self, image_path):
        result = self.call_gpt4_model_for_analysis(image_path)
        return result

    def simplify_explanation(self, result):
        simplified_explanation = self.chat_eli(result)
        return simplified_explanation

# Example usage:
if __name__ == "__main__":
    analyzer = MedicalImageAnalyzer()
    
    # Example image path
    image_path = "path_to_your_image.jpg"
    
    # Analyze the image
    result = analyzer.analyze_image(image_path)
    print("Analysis Result:")
    print(result)
    
    # Get simplified explanation
    simplified_explanation = analyzer.simplify_explanation(result)
    print("\nSimplified Explanation:")
    print(simplified_explanation)


from openai import OpenAI
import PyPDF2 as pdf
import json
from dotenv import load_dotenv
import os

# Load environment variables
load_dotenv()

class ResumeAnalyzer:
    def __init__(self, api_key):
        self.client = OpenAI(api_key=api_key)
        self.model = "gpt-4o-mini"  

    def extract_text_from_pdf(self, pdf_file):
        """Extract text from a PDF file."""
        try:
            reader = pdf.PdfReader(pdf_file)
            text = ""
            for page in reader.pages:
                text += page.extract_text()
            return text
        except Exception as e:
            raise Exception(f"Error reading PDF file: {e}")

    def analyze_resume(self, resume_text, job_description):
        """Analyze the resume based on the job description using OpenAI."""
        try:
            prompt = f"""
            You are an ATS system. Analyze this resume and return a JSON object.
            Format your response as a raw JSON object without any additional text or formatting.
            Use exactly this structure:
            {{
                "JD Match": "X%",
                "MissingKeywords": [],
                "Profile Summary": "",
                "Suggestions": []
            }}

            Resume:
            {resume_text}

            Job Description:
            {job_description}
            """

            response = self.client.chat.completions.create(
                model=self.model,
                messages=[
                    {"role": "system", "content": "You are an ATS system. Respond only with valid JSON."},
                    {"role": "user", "content": prompt},
                ],
                temperature=0.5,
                max_tokens=1000,
            )

            content = response.choices[0].message.content.strip()
            
            # Debug logging
            print("Raw API response:", content)
            
            try:
                parsed_json = json.loads(content)
                # Validate required fields
                required_fields = ["JD Match", "MissingKeywords", "Profile Summary", "Suggestions"]
                if all(field in parsed_json for field in required_fields):
                    return parsed_json
                else:
                    raise ValueError("Missing required fields in JSON response")
            except (json.JSONDecodeError, ValueError) as e:
                print(f"JSON parsing error: {str(e)}")
                return {
                    "JD Match": "0%",
                    "MissingKeywords": [],
                    "Profile Summary": "Error: Could not analyze resume",
                    "Suggestions": ["Error processing the resume. Please try again."]
                }
                
        except Exception as e:
            print(f"Analysis error: {str(e)}")
            raise Exception(f"Error analyzing resume: {e}")

# Example usage
if __name__ == "__main__":
    analyzer = ResumeAnalyzer()

    # Load job description and resume
    job_description = """
    We are looking for a software engineer with experience in Python, machine learning, and cloud computing.
    The ideal candidate should have strong problem-solving skills, experience with REST APIs, and familiarity with Docker and Kubernetes.
    """

    resume_path = "path/to/resume.pdf"  # Replace with the actual path to your resume

    try:
        # Extract text from the resume
        with open(resume_path, "rb") as pdf_file:
            resume_text = analyzer.extract_text_from_pdf(pdf_file)

        # Analyze the resume
        result = analyzer.analyze_resume(resume_text, job_description)
        print(result)  # Print the result as a dictionary
    except Exception as e:
        print(f"Error: {e}")


from langchain_openai import ChatOpenAI
from langchain.agents import AgentExecutor, create_openai_tools_agent
from langchain import hub
from langchain.tools import StructuredTool


class ResearchAgent:
    def __init__(self, api_key=None) -> None:
        # Create Tool objects with proper function references
        web_tool = StructuredTool.from_function(
            name="website_research",  # Changed from "Website Research" to "website_research"
            func=extract_relevant_sections_from_website,
            description=("Extract relevant information from a website based on keywords. "
            "Args: url (str): The URL of the website to analyze., "
            "keywords (list): A comprehesive list of keywords to search in the website based on the topic(more than 10). "
            "Returns: dict: A dictionary of relevant sections from the website."
            )
        )
        yt_tool = StructuredTool.from_function(
            name="youtube_transcript",  # Changed from "YouTube Transcript" to "youtube_transcript"
            func=youtube_transcript_loader,
            description="Extract transcript from a YouTube video"
        )
        video_to_audio_tool = StructuredTool.from_function(
            name="video_to_audio",  # Changed from "Video to Audio" to "video_to_audio"
            func=extract_audio_from_video,
            description="Extract audio from a video file"
        )
        audio_to_text_tool = StructuredTool.from_function(
            name="audio_to_text",  # Changed from "Audio to Text" to "audio_to_text"
            func=transcribe_audio,
            description="Transcribe audio to text"
        )
        self.tools = [web_tool, yt_tool, video_to_audio_tool, audio_to_text_tool]
        # self.functions = [convert_to_openai_function(t) for t in self.tools]



        # self.llm = ChatOpenAI(api_key=api_key).bind_tools(tools=self.tools)
        self.llm = ChatOpenAI(model="gpt-4o-mini", api_key=api_key).bind_tools(self.tools)

        


        # self.video_formats = ["mp4", "avi", "mkv", "mov", "flv", "wmv", "webm", "mpeg", "mpg", "3gp", "m4v", "mxf", "vob", "ogv"]
        # self.audio_formats = ["mp3", "wav", "aac", "flac", "ogg", "m4a"]

    def research_website(self, topic, url): 
        prompt1 = (
            f"Gather relavent information about topic from the website. "
            f"\nTopic: {topic} "
            f"\nWebsite: {url} "
        )
        # # context = self.llm.invoke(prompt1)
        # prompt = hub.pull("hwchase17/openai-functions-agent")
        # # Create an agent
        # agent = create_openai_tools_agent(self.llm, self.tools, prompt)

        # # Create an agent executor
        # agent_executor = AgentExecutor(agent=agent, tools=self.tools, verbose=True)

        # context = agent_executor.invoke({"input": prompt1})

        # return context['output']
        res = self.llm.invoke(prompt1)
        print(res.tool_calls[0])
        return extract_relevant_sections_from_website(**res.tool_calls[0]['args'])
    
    def extract_transcript_from_yt_video(self, url):
        prompt1 = f"Extract the text content from the youtube video. video url: {url}"
        # # context = self.llm.invoke(prompt1)
        # prompt = hub.pull("hwchase17/openai-functions-agent")
        # # Create an agent
        # agent = create_openai_tools_agent(self.llm, self.tools, prompt)

        # # Create an agent executor
        # agent_executor = AgentExecutor(agent=agent, tools=self.tools, verbose=True)

        # context = agent_executor.invoke({"input": prompt1})

        # return context['output']
        res = self.llm.invoke(prompt1)
        print(res.tool_calls[0])
        return youtube_transcript_loader(**res.tool_calls[0]['args'])

    
    def _video_to_text(self, video_path):
        audio_path = self.llm.invoke(f"Extract audio fromt the video. video path: {video_path}", return_tool_output=True)[0]
        print(audio_path)
        text = self.llm.invoke(f"Transcribe the audio into text. audio_path: {audio_path}", return_tool_output=True)[0]
        return text

    def _audio_to_text(self, audio_path):
        text = self.llm.invoke(f"Transcribe the audio into text. audio_path: {audio_path}", return_tool_output=True)[0]
        return text
    
    def extract_text_from_audio_or_video(self, file_path):
        file_format = 'Audio file' if file_path.endswith('mp3') else 'Video file'
        if file_format == 'Video file':
            text = self._video_to_text(file_path)
        elif file_format == 'Audio file':
            text = self._audio_to_text(file_path)
        return text
    

def extract_sections(url):
    import requests
    from bs4 import BeautifulSoup

    response = requests.get(url)
    soup = BeautifulSoup(response.content, 'html.parser')
    
    sections = []
    for link in soup.find_all('a', href=True):
        sections.append({
            'text': link.get_text().strip(),
            'url': link['href']
        })
        
    return sections

def filter_relevant_sections(sections, keywords):
    relevant_sections = []
    for section in sections:
        if any(keyword.lower() in section['text'].lower() for keyword in keywords):
            relevant_sections.append(section)
    
    return relevant_sections

def filter_youtube_links(sections):
    youtube_sections = []
    for section in sections.copy():  # Use copy to avoid modifying during iteration
        if 'youtube' in section['url']:
            youtube_sections.append(section)
    return youtube_sections

def gather_info_from_sections(relevant_sections):
    import requests
    from bs4 import BeautifulSoup

    content = {}
    for section in relevant_sections:
        try:
            response = requests.get(section['url'])
            soup = BeautifulSoup(response.content, 'html.parser')
            clean_text = clean_scraped_text(soup.get_text())
            content[section['url']] = clean_text
        except Exception as e:
            # print(e)
            pass
    
    return content

def clean_scraped_text(text):
    import re

    text = re.sub(r'\n+', '\n', text)
    text = re.sub(r'\s+', ' ', text)

    patterns = [
        r'Home\s+About Us.*?\s+Contact Us',
        r'This website uses cookies.*?Privacy & Cookies Policy',  
        r'Copyright.*?Powered by.*',  
    ]

    for pattern in patterns:
        text = re.sub(pattern, '', text, flags=re.DOTALL | re.IGNORECASE)

    text = re.sub(r'\|.*?\|', '', text)  
    text = text.strip()  

    return text

def youtube_transcript_loader(url):
    from youtube_transcript_api import YouTubeTranscriptApi
    try:
        video_id = url.split('/')[-1].split('=')[-1]
        transcript_list = YouTubeTranscriptApi.list_transcripts(video_id)
        transcript = transcript_list.find_transcript(['en'])

        if transcript is None:
            raise ValueError('No English transcript found for video: {}'.format(video_id))

        list_t =  transcript.fetch()
    
        transcript_text = ""
        for transcript in list_t:
            transcript_text += transcript.text+ " "
        return transcript_text
    except Exception as e:
        return f"Error: {e}"
        # raise 
    
def gather_youtube_data(sections, keywords):

    youtube_sections = []
    for i, section in enumerate(sections):
        if 'youtube' in section['url']:
            youtube_sections.append(section)

    content = {}
    for section in youtube_sections:
        text = youtube_transcript_loader(section['url'])
        if text is not None:
            content[section['url']] = text

    relevant_content = {}
    for k, v in content.items():
        if any(keyword.lower() in v.lower() for keyword in keywords):
            relevant_content[k] = v

    return relevant_content

def extract_relevant_sections_from_website(url, keywords):
    try:
        sections = extract_sections(url)
        filtered_sections = filter_relevant_sections(sections, keywords)
        gathered_info = gather_info_from_sections(filtered_sections)
        youtube_info = gather_youtube_data(sections, keywords)
        total_info = gathered_info | youtube_info
        refined_info = {url: text for url, text in total_info.items() if len(text) > 200}  # Example threshold for content length
        return refined_info
    except Exception as e:
        # return {"error": str(e)}
        raise

def extract_audio_from_video(video_path):
    from moviepy.editor import VideoFileClip
    import tempfile
    print(video_path)
    with VideoFileClip(video_path) as video:
        audio = video.audio
        temp_audio_file = tempfile.NamedTemporaryFile(delete=False, suffix=".mp3")
        audio.write_audiofile(temp_audio_file.name)
    return temp_audio_file.name

def transcribe_audio(audio_file_path):
    from openai import OpenAI
    client = OpenAI()
    with open(audio_file_path, "rb") as audio_file:
        transcription = client.audio.transcriptions.create(
            model="whisper-1",
            file=audio_file
        )   
    return transcription.text


import requests
from opencage.geocoder import OpenCageGeocode
import openai
import re
from wyge.models.openai_models import ChatOpenAI


# Function to fetch latitude and longitude for the destination using OpenCage SDK
def fetch_lat_long(api_key, location):
    geocoder = OpenCageGeocode(api_key)
    results = geocoder.geocode(location)
    if results:
        lat = results[0]['geometry']['lat']
        lon = results[0]['geometry']['lng']
        return lat, lon
    return None, None


# Function to fetch real-time weather data
def fetch_weather(api_key, lat, lon):
    url = f"https://api.openweathermap.org/data/2.5/weather?lat={lat}&lon={lon}&appid={api_key}"
    response = requests.get(url)
    if response.status_code == 200:
        weather_data = response.json()
        print(weather_data)
        return {
            "temperature": weather_data["main"]["temp"] - 273.15,  # Convert Kelvin to Celsius
            "condition": weather_data["weather"][0]["description"].capitalize()
        }
    return "Weather data not available."


# TravelPlannerAgent class that generates a travel plan based on user input
class TravelPlannerAgent:
    def __init__(self, weather_api_key, geolocation_api_key, openai_api_key):
        self.weather_api_key = weather_api_key
        self.geolocation_api_key = geolocation_api_key
        self.llm = ChatOpenAI(api_key=openai_api_key)  # Initialize OpenAI LLM

    def parse_user_input(self, prompt):
        # Extract destination and duration (days) from the prompt
        match = re.search(r'travel plan to (\w+)(?: for (\d+) days)?', prompt, re.IGNORECASE)
        destination = match.group(1)
        days = int(match.group(2)) if match.group(2) else 3  # default to 3 days if not specified
        return destination, days

    def gather_destination_info(self, destination):
        # Get latitude and longitude of the destination
        lat, lon = fetch_lat_long(self.geolocation_api_key, destination)
        if lat is None or lon is None:
            return "Unable to retrieve location data."

        # Fetch real-time weather data for the specified destination using latitude and longitude
        weather = fetch_weather(self.weather_api_key, lat, lon)

        return {
            "lat": lat,
            "lon": lon,
            "weather": weather,
        }

    def generate_itinerary(self, destination, days):
        # Gather real-time context data
        context = self.gather_destination_info(destination)
        print("context is ...................")
        print(context)
        if isinstance(context, str):
            return context  # Return error message if location data is unavailable

        # LLM prompt with a structured request
        prompt = (
            # f"Create a comprehensive travel itinerary for {destination} over {days} days. "
            # f"Include local attractions, historical sites, and unique activities. "
            # f"Structure each day with 'Morning,' 'Afternoon,' and 'Evening' sections, "
            # f"and use bullet points to list the activities for each part of the day. "
            # f"Provide brief descriptions for each place, mentioning its significance or unique qualities. "
            # f"Include weather-appropriate recommendations based on current weather conditions: "
            # f"Temperature: {context['weather']['temperature']}°C, Condition: {context['weather']['condition']}. "
            # f"Optimize for cost-effective options where possible.\n\n"
            # f"Format the itinerary in markdown with clear headers, bullet points, and sections. "
            # f"Highlight important tips, and end with a short summary of the trip."

            f"Create a comprehensive travel itinerary for {destination} over {days} days. "
            f"Provide brief description  and importance of {destination}, mentioning its significance or unique qualities. "
            f"Include recommendations for local attractions, historical sites, and unique activities, optimizing for the region's latitude and longitude, and adapting plans based on weather conditions. "
            f" Do not structure each day like morning,afternoon,evening sections, and use bullet points to list activities for each day. "
            f"Suggest suitable accommodations based on the location's latitude and longitude, factoring in weather conditions and accessibility to major attractions. "
            f"Adapt food and accommodation suggestions based on temperature ({context['weather']['temperature']}°C) and condition ({context['weather']['condition']}), "
            f"including ideas like indoor dining spots for rainy days or places with outdoor seating in pleasant weather."
            f"Offer helpful travel tips tailored to the location, such as recommended modes of transport (e.g., metro, bikes, or taxis), nearby transportation hubs, and local etiquette. "
            f"Highlight any important safety tips or local customs that visitors should be aware of to enhance their experience and avoid cultural misunderstandings. "
            f"Provide practical packing tips based on the weather, suggesting items like umbrellas, sunscreen, or specific clothing for comfort. "
            f"Format the itinerary in markdown with clear headers, bullet points, and sections. "
            f"Recommend the famous foods that reflect the destination’s culture and cuisine. "
            # f"End with a short summary of the trip and highlight any essential tips or recommendations for a successful journey."

        )
        print("llm going")
        # Use OpenAI to generate itinerary
        response = self.llm.run(prompt)
        print("______________________________________________________")
        return response

    def generate_travel_plan(self, prompt):
        # Parse prompt and create the travel plan
        destination, days = self.parse_user_input(prompt)
        itinerary = self.generate_itinerary(destination, days)

        travel_plan = f"### Travel Itinerary for {destination} - {days} Days\n\n{itinerary}"
        return travel_plan


if __name__ == "__main__":

    # API keys for the services
    weather_api_key = 'f701a9b1299fa3bbb07471570c730090'
    geolocation_api_key = '2f85379af3084ad1a9fc724dfa71b041'
    openai_api_key = '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'

    # Initialize the TravelPlannerAgent with the API keys
    agent = TravelPlannerAgent(weather_api_key, geolocation_api_key, openai_api_key)

    # Example prompt for generating a travel plan
    prompt = "I need a travel plan to Paris for 5 days"

    # Generate the travel plan
    travel_plan = agent.generate_travel_plan(prompt)

    # Print the generated travel plan
    print(travel_plan)


from setuptools import setup, find_packages

setup(
    name="wyge",
    version="1.1.7",
    packages=find_packages(),
    package_dir={"": "."},
    install_requires=['openai', 'requests', 'numpy', 'pydantic[email]', 'langchain'], 
    author="Prudvi",
    author_email="prudhvisneha2003@gmail.com",
    description="A Python library for WYGE",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/prudvireddyNS/vyzeai",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    extras_require={
        "dev": ["pytest>=7.0", "twine>=4.0.2"],
    },
    python_requires=">=3.8",
)


import os
from typing import Dict, List, Any
from dotenv import load_dotenv
from langchain import PromptTemplate

# Load environment variables
load_dotenv()
api_key = os.getenv("OPENAI_API_KEY")

# Define the instructor prompt template
INSTRUCTOR_PROMPT = """
As a teaching assistant, your task is to teach the user based on the provided syllabus.
The syllabus outlines the specific topics, concepts, and learning objectives to be covered.

Syllabus for {topic}:
{syllabus}

Follow the topics in order and provide comprehensive instruction to convey the knowledge to the user.
Maintain a supportive and approachable demeanor, creating a positive learning environment.

Previous conversation:
{conversation_history}

Provide a clear, concise response to continue the lesson.
"""

def create_teaching_agent(llm: Any) -> Dict:
    """
    Create a teaching agent that can provide educational responses based on a syllabus.
    
    Args:
        llm: The language model to use for generating responses
        
    Returns:
        Dict: A dictionary containing the teaching agent state and functions
    """
    # Initialize the agent state
    agent_state = {
        "syllabus": "",
        "topic": "",
        "conversation_history": []
    }
    
    # Create a function to seed the agent with a syllabus and topic
    def seed(syllabus: str, topic: str) -> None:
        agent_state["syllabus"] = syllabus
        agent_state["topic"] = topic
        agent_state["conversation_history"] = []
    
    # Create a function to add a user message to the conversation history
    def add_user_message(message: str) -> None:
        agent_state["conversation_history"].append(f"User: {message}")
    
    # Create a function to generate a response from the agent
    def generate_response() -> str:
        # Create the prompt template
        prompt_template = PromptTemplate(
            template=INSTRUCTOR_PROMPT,
            input_variables=["syllabus", "topic", "conversation_history"]
        )
        
        # Format the prompt with the current state
        prompt = prompt_template.format(
            syllabus=agent_state["syllabus"],
            topic=agent_state["topic"],
            conversation_history="\n".join(agent_state["conversation_history"])
        )
        
        # Generate the response using the provided LLM
        response = llm.invoke(prompt)
        
        # Add the response to the conversation history
        agent_state["conversation_history"].append(f"Assistant: {response.content}")
        
        return response
    
    # Add a function to get conversation history
    def get_conversation_history() -> List[str]:
        return agent_state["conversation_history"]
    
    # Return the agent interface
    return {
        "seed": seed,
        "add_user_message": add_user_message,
        "generate_response": generate_response,
        "conversation_history": get_conversation_history  # Add this line
    }

def teaching_agent_fun(llm: Any) -> Dict:
    """
    Factory function to create a teaching agent.
    
    Args:
        llm: The language model to use for generating responses
        
    Returns:
        Dict: A dictionary containing the teaching agent functions
    """
    return create_teaching_agent(llm)


import os
import json
import time
from typing import Optional, List, Dict, Any, Type, Union

import openai
from pydantic import BaseModel

from wyge.memory.basic_memory import MemoryManager
from wyge.tools.base_tool import ToolCall, execute_tool_calls, format_tool_for_api, ResponseParser
from wyge.common.logger import logger, logging

class ChatOpenAI:
    """Enhanced OpenAI chat model wrapper with advanced features"""
    
    def __init__(
        self,
        api_key: Optional[str] = None,
        model_name: str = 'gpt-4o-mini',  
        temperature: float = 0.2,
        max_tokens: int = 1500,
        tools: Optional[List[Type[BaseModel]]] = None,
        memory_enabled: bool = False,
        memory_size: int = 20,
        default_system_message: str = "You are a helpful assistant.",
        retry_count: int = 3,
        retry_delay: float = 1.0,
        verbose: bool = False
    ):
        # API setup
        self.api_key = api_key or os.getenv('OPENAI_API_KEY')
        if not self.api_key:
            raise ValueError("API key must be provided either via argument or 'OPENAI_API_KEY' environment variable")
        os.environ['OPENAI_API_KEY'] = self.api_key
        
        # Model configuration
        self.model_name = model_name
        self.temperature = temperature
        self.max_tokens = max_tokens
        self._validate_model_config()
        
        # Tools setup
        self.tools = []
        if tools:
            self.register_tools(tools)
        
        # Memory setup
        self.memory_enabled = memory_enabled
        self.memory = MemoryManager() 
        if memory_enabled:
            self.memory = MemoryManager(
                max_size=memory_size,
                default_system_message=default_system_message
            )
        
        # Other settings
        self.retry_count = retry_count
        self.retry_delay = retry_delay
        self.verbose = verbose
        
        # Tracking
        self.usage_stats = {
            'total_tokens': 0,
            'prompt_tokens': 0,
            'completion_tokens': 0,
            'successful_calls': 0,
            'failed_calls': 0,
            'total_cost': 0.0  # Estimated
        }
        
        if verbose:
            logger.setLevel(logging.DEBUG)
        
    def _validate_model_config(self) -> None:
        """Validate model configuration parameters"""
        if not 0 <= self.temperature <= 2:
            raise ValueError("Temperature must be between 0 and 2")
        if self.max_tokens < 1:
            raise ValueError("max_tokens must be positive")
    
    def register_tools(self, tool_models: List[Type[BaseModel]]) -> None:
        """Register tools with the model"""
        for model in tool_models:
            tool_config = format_tool_for_api(model)
            self.tools.append(tool_config)
            logger.debug(f"Registered tool: {tool_config['function']['name']}")
    
    def update_system_message(self, content: str) -> None:
        """Update the system message"""
        self.memory.update_system_message(content)
        if self.memory_enabled:
            self.memory.update_system_message(content)
    
    def _build_messages(self, prompt: Optional[str], system_message: Optional[str]) -> List[Dict[str, Any]]:
        """Build the messages list for the API call"""
        if self.memory_enabled:
            if system_message:
                self.memory.update_system_message(system_message)
            
            if prompt:
                should_summarize = self.memory.add_message('user', prompt)
                if should_summarize:
                    self._summarize_memory()
                    
            return self.memory.get_conversation_history()
        else:
            # Simple message format when memory is disabled
            return [
                {"role": "system", "content": system_message or "You are a helpful assistant."},
                {"role": "user", "content": prompt or ""}
            ]
    
    def _summarize_memory(self) -> None:
        """Summarize the memory using a separate LLM call"""
        if not self.memory_enabled:
            return
            
        def summarizer(messages: List[Dict[str, Any]]) -> str:
            # Convert messages to a string format
            conversation = "\n".join([
                f"{msg['role']}: {msg['content']}" 
                for msg in messages 
                if 'content' in msg and msg['content']
            ])
            
            # Create a summarization prompt
            prompt = (
                "Summarize the following conversation, preserving key information, "
                "numeric data, and the sequence of events. Be concise but thorough:\n\n"
                f"{conversation}"
            )
            
            # Make a new API call just for summarization
            try:
                response = openai.chat.completions.create(
                    model=self.model_name,
                    messages=[
                        {"role": "system", "content": "You are a helpful summarization assistant."},
                        {"role": "user", "content": prompt}
                    ],
                    temperature=0.3,
                    max_tokens=500
                )
                
                if response.choices and response.choices[0].message.content:
                    return response.choices[0].message.content
                return "Failed to generate summary."
            except Exception as e:
                logger.error(f"Summarization failed: {str(e)}")
                return "Conversation summary unavailable due to an error."
        
        self.memory.summarize(summarizer)
    
    def _call_api(self, messages: List[Dict[str, Any]], 
                return_tool_output: bool = False) -> Dict[str, Any]:
        """Make the API call with retry logic"""
        for attempt in range(self.retry_count):
            try:
                if self.verbose:
                    logger.debug(f"API call attempt {attempt+1}/{self.retry_count}")
                    logger.debug(f"Messages: {json.dumps(messages, indent=2)}")
                
                response = openai.chat.completions.create(
                    model=self.model_name,
                    messages=messages,
                    temperature=self.temperature,
                    max_tokens=self.max_tokens,
                    tools=self.tools if self.tools else None
                )
                print("Raw Response: ", response)
                
                # Update usage statistics
                if hasattr(response, 'usage'):
                    self.usage_stats['total_tokens'] += response.usage.total_tokens
                    self.usage_stats['prompt_tokens'] += response.usage.prompt_tokens
                    self.usage_stats['completion_tokens'] += response.usage.completion_tokens
                    # TODO: Add cost estimation based on model pricing
                
                self.usage_stats['successful_calls'] += 1
                
                # Parse the response
                parsed = ResponseParser.parse_completion(response)
                
                if self.verbose:
                    try:
                        logger.debug(f"Response: {json.dumps(parsed, indent=2, cls=ToolCallEncoder)}")
                    except Exception as e:
                        logger.warning(f"Could not serialize response for logging: {str(e)}")
                
                # Check if we need to execute tool calls
                tool_results = None
                
                if parsed['finish_reason'] in ('function_call', 'tool_calls') and parsed['tool_calls']:
                    tool_results = execute_tool_calls(parsed['tool_calls'])
                    
                    if self.verbose:
                        logger.debug(f"Tool results: {json.dumps([t.to_dict() for t in tool_results], indent=2)}")
                    
                    # Add tool results to memory if enabled
                    if self.memory_enabled:
                        # First add the assistant's message with tool calls
                        self.memory.messages.append(response.choices[0].message.to_dict())
                        
                        # Then add each tool result
                        for tool_call in tool_results:
                            self.memory.add_tool_message(tool_call)
                
                return {
                    'response': response,
                    'parsed': parsed,
                    'tool_results': tool_results
                }
                
            except Exception as e:  # Use normal error handling
                self.usage_stats['failed_calls'] += 1
                logger.error(f"API call error (attempt {attempt+1}/{self.retry_count}): {str(e)}")
                
                if attempt < self.retry_count - 1:
                    time.sleep(self.retry_delay * (2 ** attempt))  # Exponential backoff
                else:
                    raise Exception(f"API call failed after {self.retry_count} attempts: {str(e)}")
                
    def run(self, 
            prompt: Optional[str] = None, 
            system_message: Optional[str] = None,
            return_tool_output: bool = False) -> Union[str, List[ToolCall]]:
        """
        Run the model with the given prompt
        
        Args:
            prompt: The user message
            system_message: Optional system message override
            return_tool_output: If True, returns tool results instead of final response
            
        Returns:
            Either the model's response text or tool results depending on return_tool_output
        """
        # Build messages
        messages = self._build_messages(prompt, system_message)
        
        # Make the initial API call
        result = self._call_api(messages, return_tool_output)
        
        # If we need to return tool output directly
        if return_tool_output and result['tool_results']:
            return result['tool_results']
        
        # If we have tool results, make another call to get the final response
        if result['tool_results'] and not self.memory_enabled:
            # Add the assistant's response with tool calls
            messages.append(result['response'].choices[0].message.to_dict())
            
            # Add tool results
            for tool_call in result['tool_results']:
                messages.append({
                    'role': 'tool',
                    'content': json.dumps(tool_call.to_dict()),
                    'tool_call_id': tool_call.id
                })
            
            # Make another API call for the final response
            result = self._call_api(messages)
        
        # Get the final content
        content = result['parsed']['content']
        
        return content
    
    def clear_memory(self, keep_system_message: bool = True) -> None:
        """Clear the conversation memory"""
        if self.memory_enabled:
            self.memory.clear(preserve_system=keep_system_message)
    
    def search_memory(self, query: str, exact_match: bool = False) -> List[Dict[str, Any]]:
        """Search for messages in memory"""
        if self.memory_enabled:
            return self.memory.search(query, exact_match=exact_match)
        return []
    
    def last_message(self) -> Optional[Dict[str, Any]]:
        """Get the last message in memory"""
        if self.memory_enabled and self.memory.messages:
            return self.memory.messages[-1]
        return None
    
    def get_usage_stats(self) -> Dict[str, Any]:
        """Get usage statistics"""
        return self.usage_stats
    
    def get_config(self) -> Dict[str, Any]:
        """Get the current configuration"""
        # Mask API key for security
        masked_api_key = '*' * (len(self.api_key) - 4) + self.api_key[-4:] if self.api_key else None
        
        return {
            'api_key': masked_api_key,
            'model_name': self.model_name,
            'temperature': self.temperature,
            'max_tokens': self.max_tokens,
            'memory_enabled': self.memory_enabled,
            'tools_count': len(self.tools),
            'retry_settings': {
                'count': self.retry_count,
                'delay': self.retry_delay
            },
            'verbose': self.verbose
        }
    
class ToolCallEncoder(json.JSONEncoder):
    def default(self, obj):
        if hasattr(obj, '__dict__'):
            return obj.__dict__
        return str(obj)


from langchain_community.document_loaders import TextLoader
from langchain.text_splitter import RecursiveCharacterTextSplitter
from langchain_community.embeddings import OpenAIEmbeddings
from langchain_community.vectorstores import FAISS
from langchain.prompts import ChatPromptTemplate
from langchain_community.chat_models import ChatOpenAI
from langchain.schema.runnable import RunnablePassthrough
from langchain.schema.output_parser import StrOutputParser
from langchain.schema import Document
from PyPDF2 import PdfReader

class CustomPDFLoader:
    def __init__(self, file_path, page_range=None):
        self.file_path = file_path
        self.page_range = page_range

    def load(self):
        reader = PdfReader(self.file_path)
        text = ""
        
        if self.page_range:
            start_page, end_page = self.page_range
            for i, page in enumerate(reader.pages):
                if i >= start_page and i <= end_page:
                    text += page.extract_text() + "\n"
        else:
            for page in reader.pages:
                text += page.extract_text() + "\n"
        
        return [Document(page_content=text, metadata={"source": self.file_path, "page_range": self.page_range})]


class RAGApplication:
    def __init__(self, file_paths, openai_api_key, model_name="gpt-3.5-turbo", chunk_size=500, chunk_overlap=50, page_range=None):
        self.page_range = page_range
        self.file_paths = file_paths
        self.openai_api_key = openai_api_key
        self.model_name = model_name
        self.chunk_size = chunk_size
        self.chunk_overlap = chunk_overlap
        self.embeddings = OpenAIEmbeddings(openai_api_key=self.openai_api_key)
        self.vectorstore = None
        self.retriever = None
        self.rag_chain = None
        self._initialize_components()

    def _initialize_components(self):
        text_chunks = []

        # Load and split each document
        for file_path in self.file_paths:
            if file_path.endswith('.pdf'):
                loader = CustomPDFLoader(file_path, self.page_range)
            else:
                loader = TextLoader(file_path, encoding="utf8")

            document = loader.load()
            text_splitter = RecursiveCharacterTextSplitter(chunk_size=self.chunk_size, chunk_overlap=self.chunk_overlap)
            chunks = text_splitter.split_documents(document)
            text_chunks.extend(chunks)

        # Create vector store and retriever
        self.vectorstore = FAISS.from_documents(text_chunks, self.embeddings)
        self.retriever = self.vectorstore.as_retriever()

        # Set up the RAG chain
        template = """You are an assistant for question-answering tasks.
        Use the following pieces of retrieved context to answer the question.
        If you don't know the answer, just say that you don't know.
        Use ten sentences maximum and keep the answer concise.
        Question: {question}
        Context: {context}
        Answer:
        """
        prompt = ChatPromptTemplate.from_template(template)
        llm_model = ChatOpenAI(openai_api_key=self.openai_api_key, model_name=self.model_name)
        output_parser = StrOutputParser()

        self.rag_chain = (
            {"context": self.retriever, "question": RunnablePassthrough()}
            | prompt
            | llm_model
            | output_parser
        )

    def query(self, question):
        return self.rag_chain.invoke(question)

# Example usage:
if __name__ == "__main__":
    rag_app = RAGApplication(file_paths=["C:/Users/prudh/Desktop/Resume.pdf"], openai_api_key="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")
    print(rag_app.query("what is my name?"))
    print(rag_app.query("What are my skills?"))
